/*
 * Author: Pramod R. Sanaga (pramod@cs.utah.edu)
 * Module to translate XQuery to SQL
 * ----------------------------------
 * Class Project - CS6530 - Fall '08
 * 
 * Utilizes the Shrex XPath -> SQL query translation and XML -> Relation mapping
 * 
 * Accepts: Any XQuery expression containing only the XPath expressions accepted by Shrex XPath query
 * translation module.
 * .
 * Output: Evaluated output of the XQuery expression(s)
 * 
 * 
 */

package gui;

import edu.ogi.cse.xs.database.DBConnection;
import edu.ogi.cse.xs.database.DataBase;
import edu.ogi.cse.xs.query.SQLQuery;
import edu.ogi.cse.xs.query.XPathQuery;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.sql.ResultSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.*;
import java.io.*;

import net.sf.saxon.Configuration;
import net.sf.saxon.query.DynamicQueryContext;
import net.sf.saxon.query.StaticQueryContext;
import net.sf.saxon.query.XQueryExpression;
import net.sf.saxon.trans.XPathException;

import javax.xml.transform.*;
import javax.xml.transform.stream.StreamResult;
import java.util.Properties;

public class XQueryPanel extends JSplitPane {

	private JTextArea m_queryArea, m_resultArea;
	private JButton m_buttonExecute;
	private XPathQuery m_translator;

	// Code to create the Swing GUI for the Xquery panel.
	public XQueryPanel() {
		super(JSplitPane.VERTICAL_SPLIT);

		m_queryArea = new JTextArea();
		m_queryArea.setEditable(true);
		m_queryArea.setBackground(Color.white);

		m_resultArea = new JTextArea();
		m_resultArea.setEditable(false);
		m_resultArea.setBackground(Color.white);

		m_buttonExecute = new JButton("Execute query");
		m_buttonExecute.addActionListener(new ExecuteAction());

		setDividerSize(4);
		setDividerLocation(200);

		JPanel topPanel = new JPanel();

		topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.PAGE_AXIS));

		topPanel.add(new JLabel("Enter XQuery expression below:"));

		// Adds a text area for entering the query.
		topPanel.add(new JScrollPane(m_queryArea));

		// This button executes the query and prints the result
		topPanel.add(m_buttonExecute);

		setLeftComponent(topPanel);

		// Destination text area for the query result.
		setRightComponent(new JScrollPane(m_resultArea));

		// Keep a reference to the Shrex schema translator
		m_translator = new XPathQuery(SchemaPanel.getInstance().getMapping());
	}

	// Consults Shrex's schema translator to get a translated list of SQL queries
	// corresponding to an XPath expression.
	private java.util.List GetSqlQueries(String XPathExpr) {
		return m_translator.translate(XPathExpr);
	}

	// Connects to a database connection, evaluates the given SQL queries and returns
	// the results.
	private String EvalSqlQuery(java.util.List sqlQueries,
			DBConnection selectedConn) {

		if (selectedConn != null) {
			try {
				// Create a String from the List of sqlQueries passed as a parameter.
				String sqlstr = "";

				for (int j = 0; j < sqlQueries.size(); j++) {
					SQLQuery sql = (SQLQuery) sqlQueries.get(j);

					sqlstr += sql.getQueryString() + "\n";
					if (j < sqlQueries.size() - 1) {
						sqlstr += "UNION ALL" + "\n";
					}
				}

				// Execute the SQL queries on the database connection selected by the user.
				ResultSet rs = selectedConn.executeQuery(sqlstr);

				return m_translator.dumpResultSet(rs, sqlQueries);
			} catch (Throwable t) {
				t.printStackTrace();
			}
		}

		// If there is no current database connection, return nothing.
		return "";
	}

	// Delete a non-empty directory - used for removing temporary files.
	void DeleteDirectory(String dirName) {
		File dirHandle = new File(dirName);

		if (dirHandle.exists()) {
			File[] filesList = dirHandle.listFiles();

			for (int i = 0; i < filesList.length; i++)
				filesList[i].delete();
		}

	}

	class ExecuteAction extends AbstractAction {

		ExecuteAction() {
			super("Execute query");
		}

		public void actionPerformed(ActionEvent e) {

			XSGui xsgui = XSGui.getInstance();

			DataBase db = xsgui.getDB();
			java.util.List connections = db.connections();

			if (connections == null || connections.size() <= 0) {
				String message = "No Database Conneciton is available. Please add a connection first.";
				JOptionPane.showMessageDialog(xsgui, message);
			} else {

				// Ask the user to select a database connection.
				DBConnection selectedConn = xsgui.promptForAConnection(connections, "Evaluate SQL Query");

				String queryText = m_queryArea.getText();

				// Regular expression to detect absolute XPath expressions in the given query.
				Pattern pathExprPattern = Pattern
						.compile("([\\s]|(^\\$\\w))((//|/)[\\w\\d\\*\\[\\]]*)+");
				
				// Create a matcher object with the above regex and user's query.
				Matcher pathMatcher = pathExprPattern.matcher(queryText);
				StringBuffer stringBuf = new StringBuffer();

				int tmpFileCount = 1;

				// Create a temporary directory to hold the intermediate results of the 
				// XPath expressions.
				new java.io.File(".tmp").mkdir();

				// Find absolute XPath expressions in the query and replace them with
				// references to temporary doc(".tmp/xpath$.xml")/ references.
				
				// We make use of Shrex's XPath query translation to evaluate the absolute
				// path expressions and put the results in temporary files.
				while (pathMatcher.find()) {
					
					String matchedPathExpr = pathMatcher.group();
					
					// Generate a temporary file name.
					String tmpFileName = ".tmp/xpath"
							+ (new Integer(tmpFileCount)).toString() + ".xml";

					// Get the list of SQL queries for this absolute XPath expression.
					java.util.List sqlQueries = GetSqlQueries(matchedPathExpr);
					
					// Evaluate the above SQL statements and get the XML results from the database.
					String xpathResult = EvalSqlQuery(sqlQueries, selectedConn);

					// Print the XML elements to the temporary file with a root element - to make
					// it a valid XML file.
					try {
						PrintWriter fileWriter = new PrintWriter(
								new BufferedWriter(new FileWriter(tmpFileName)));
						fileWriter.write("<RootElement>\n");
						fileWriter.write(xpathResult);
						fileWriter.write("\n</RootElement>\n");
						fileWriter.close();
					} catch (IOException ioError) {
					}

					// Replace the absolute path expression in the original query 
					// with this modified location.
					pathMatcher.appendReplacement(stringBuf, " doc(\""
							+ tmpFileName + "\")/RootElement/* ");

					tmpFileCount = tmpFileCount + 1;
				}

				pathMatcher.appendTail(stringBuf);

				// Run the modified XQuery expression(s) and display the results to the user.
				try {

					final Configuration config = new Configuration();
					final StaticQueryContext sqc = new StaticQueryContext(
							config);
					
					// Compile the XQuery expression using Saxon.
					final XQueryExpression exp = sqc.compileQuery(stringBuf
							.toString());

					final DynamicQueryContext dynamicContext = new DynamicQueryContext(
							config);
					final Properties props = new Properties();
					props.setProperty(OutputKeys.METHOD, "xml");
					props.setProperty(OutputKeys.INDENT, "yes");

					StringWriter stringWriter = new StringWriter();

					// Run the compiled XQuery and read results into a String.
					exp.run(dynamicContext, new StreamResult(stringWriter),
							props);

					// Display results.
					m_resultArea.setText(stringWriter.toString());
				} catch (XPathException xpathError) {
				}

				// Remove the temporary .xml files
				DeleteDirectory(".tmp");
			}

		}
	}
}
